/*
    UHU-menu -- input plugin for freedesktop-style files
    Copyright (C) 2003  Koblinger Egmont <egmont@uhulinux.hu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>

#include "uhu-menu.h"

static struct item *item_new (void)
{
	struct item *item;
	item = malloc(sizeof(struct item));
	item->subitems_nr = 0;
	item->subitems = NULL;
	return item;
}

static int item_add_subitem (struct item *item, const struct item *subitem)
{
	if (item->subitems_nr == 0) {
		item->subitems = malloc(sizeof(struct item));
	} else {
		item->subitems = realloc(item->subitems, (item->subitems_nr + 1) * sizeof(struct item));
	}
	if (subitem != NULL) {
		memcpy(&item->subitems[item->subitems_nr], subitem, sizeof(struct item));
	} else {
		item->subitems[item->subitems_nr].subitems_nr = 0;
		item->subitems[item->subitems_nr].subitems = NULL;
	}
	item->subitems_nr++;
	return 0;
}

static int replace_item (struct item *dst, struct item *src)
{
	src->issubmenu = dst->issubmenu;
	src->subitems = dst->subitems;
	src->subitems_nr = dst->subitems_nr;
	src->file_name = dst->file_name;
	memcpy(dst, src, sizeof(struct item));
	return 0;
}

static int insert_directory_subitem (struct item *item, struct item *subitem,
  const char *path)
{
	char buf[PATH_MAX];
	int i, j;

	buf[0] = '\0';
	path += strspn(path, "/");
	while (path[0] != '\0') {
		j = strcspn(path, "/");
		sprintf(buf + strlen(buf), "/%.*s", j, path);
		path += j;
		path += strspn(path, "/");
		for (i = 0; i < item->subitems_nr; i++) {
			if (!(strcmp(item->subitems[i].file_name, buf))) {
				break;
			}
		}
		if (i == item->subitems_nr) {
			item_add_subitem(item, NULL);
			item->subitems[i].issubmenu = 1;
			item->subitems[i].file_name = strdup(buf);
			item->subitems[i].sections = NULL;
		}
		item = &item->subitems[i];
	}
	if (item->sections != NULL) {
		message(3, "%s menu directory duplicated", buf[0] == '\0' ? "root" : buf);
		return -1;
	}
	replace_item(item, subitem);
	return 0;
}

static int insert_file_subitem (struct item *item, struct item *subitem,
  const char *path)
{
	char buf[PATH_MAX];
	int i, j;

	buf[0] = '\0';
	path += strspn(path, "/");
	while (path[0] != '\0') {
		j = strcspn(path, "/");
		sprintf(buf + strlen(buf), "/%.*s", j, path);
		path += j;
		path += strspn(path, "/");
		for (i = 0; i < item->subitems_nr; i++) {
			if (!(strcmp(item->subitems[i].file_name, buf))) {
				break;
			}
		}
		if (i == item->subitems_nr) {
			if (path[0] != '\0') {
				message(3, "unknown value for X-UHU-Menu");
				return -1;
			}
			item_add_subitem(item, subitem);
			item->subitems[i].issubmenu = 0;
			item->subitems[i].file_name = strdup(buf);
			return 0;
		}
		item = &item->subitems[i];
	}
	return -1;
}

struct item *read_menu (const char *path)
{
	struct item *rootitem;
	struct item item;
	char *fn;
	char fn2[PATH_MAX];
	char buf[PATH_MAX];
	char buf2[PATH_MAX];
	struct dirent *de;
	DIR *d;
	int l;

	rootitem = item_new();
	rootitem->issubmenu = 1;
	rootitem->file_name = "";
	rootitem->sections = NULL;

	sprintf(buf, "%s/desktop-directories", path);
	d = opendir(buf);
	if (d == NULL) {
		message_e(3, "opendir %s failed", buf);
		return NULL;
	}
	while ((de = readdir(d)) != NULL) {
		if (de->d_name[0] == '.') continue;
		l = strlen(de->d_name);
		if (l < 10 || strcmp(de->d_name + l - 10, ".directory")) continue;

		sprintf(buf2, "%s/%s", buf, de->d_name);
		if (desktop_parse(buf2, 0, &item) >= 0) {
			fn = get_value(&item, NULL, "X-UHU-Menu", "");
			if (fn != NULL) {
				insert_directory_subitem(rootitem, &item, fn);
			}
		}
	}
	closedir(d);

	sprintf(buf, "%s/applications", path);
	d = opendir(buf);
	if (d == NULL) {
		message_e(3, "opendir %s failed", buf);
		return NULL;
	}
	while ((de = readdir(d)) != NULL) {
		if (de->d_name[0] == '.') continue;
		l = strlen(de->d_name);
		if (l < 8 || strcmp(de->d_name + l - 8, ".desktop")) continue;

		sprintf(buf2, "%s/%s", buf, de->d_name);
		if (desktop_parse(buf2, 0, &item) >= 0) {
			fn = get_value(&item, NULL, "X-UHU-Menu", "");
			if (fn != NULL) {
				sprintf(fn2, "%s/%s", fn, de->d_name);
				insert_file_subitem(rootitem, &item, fn2);
			}
		}
	}
	closedir(d);

	return rootitem;
}
